from asyncio import gather
from asyncio.locks import BoundedSemaphore
from collections.abc import AsyncGenerator
from contextlib import asynccontextmanager
from dataclasses import dataclass

from aiobotocore.config import AioConfig
from botocore.client import Config
from opentelemetry.trace import get_tracer
from types_aiobotocore_s3.client import S3Client

from .aws import AWSRegion, get_bucket_region, sess

tracer = get_tracer(__name__)


@dataclass
class RegionClientList:
    sema: BoundedSemaphore
    client_add_sema: BoundedSemaphore
    clients: list[S3Client]


class AsyncS3ClientPool:
    max_clients_per_region: int | None = None
    region_clients: dict[AWSRegion, RegionClientList] | None = None

    async def open(self, max_clients_per_region: int):
        with tracer.start_as_current_span(
            "open pool", attributes={"max_clients_per_region": max_clients_per_region}
        ):
            self.max_clients_per_region = max_clients_per_region
            self.region_clients = {}

    async def close(self):
        with tracer.start_as_current_span("close pool"):
            if self.max_clients_per_region is None or self.region_clients is None:
                raise RuntimeError("Pool not open")

            close_tasks = []
            for region_client in self.region_clients.values():
                for client in region_client.clients:
                    close_tasks.append(client.__aexit__(None, None, None))

            await gather(*close_tasks)
            self.region_clients = None
            self.max_clients_per_region = None

    @asynccontextmanager
    async def s3_client_for_bucket(
        self, bucket_name: str
    ) -> AsyncGenerator[S3Client, None]:
        if self.max_clients_per_region is None or self.region_clients is None:
            raise RuntimeError("Pool not open")

        with tracer.start_as_current_span(
            "s3 client for bucket", attributes={"bucket": bucket_name}
        ) as s:
            with tracer.start_as_current_span(
                "get bucket region", attributes={"bucket": bucket_name}
            ):
                region = await get_bucket_region(Bucket=bucket_name)

            s.set_attributes({"region": region})

            if region not in self.region_clients:
                self.region_clients[region] = RegionClientList(
                    sema=BoundedSemaphore(self.max_clients_per_region),
                    client_add_sema=BoundedSemaphore(1),
                    clients=[],
                )

            region_client = self.region_clients[region]

            async with region_client.sema:
                async with region_client.client_add_sema:
                    if len(region_client.clients) == 0:
                        with tracer.start_as_current_span(
                            "create s3 client",
                            attributes={"region": region},
                        ):
                            # https://github.com/boto/boto3/issues/421 - use addressing_style=path to support buckets before DNS propagation
                            client = await sess.create_client(
                                "s3",
                                region_name=region,
                                config=AioConfig().merge(
                                    Config(
                                        signature_version="s3v4",
                                        s3={"addressing_style": "path"},
                                    )
                                ),
                            ).__aenter__()
                    else:
                        client = region_client.clients.pop()

                try:
                    yield client
                finally:
                    region_client.clients.append(client)
                    s.set_attributes(
                        {
                            f"client_count_{region}": len(region_client.clients),
                            "region_count": len(self.region_clients),
                        }
                    )


s3_pool = AsyncS3ClientPool()



import asyncio
from dataclasses import dataclass
from typing import Any

import aiohttp
from google.api_core.exceptions import ServerError
from google.auth import default
from google.auth.transport import requests


@dataclass(frozen=True)
class PubsubMessage:
    data: str | None
    attributes: dict[str, str] | None
    messageId: str
    publishTime: str
    orderingKey: str | None = None


@dataclass(frozen=True)
class SubscriptionMessage:
    ackId: str | None
    message: PubsubMessage | None
    deliveryAttempt: int | None


@dataclass(frozen=True)
class SubscriptionMessageResp:
    receivedMessages: list[SubscriptionMessage]


class AsyncPubsubClient:
    credentials, _ = default(scopes=["https://www.googleapis.com/auth/cloud-platform"])

    def __init__(
        self, session: aiohttp.ClientSession, project_id: str, subscription: str
    ):
        self.project_id = project_id
        self.subscription = subscription
        self.api_url = f"https://pubsub.googleapis.com/v1/projects/{self.project_id}/subscriptions/{self.subscription}"
        self.session = session

    @staticmethod
    def get_auth_token():
        credentials = AsyncPubsubClient.credentials

        if credentials.token is None or credentials.expired:
            credentials.refresh(requests.Request())
        return credentials.token

    async def _make_api_request(self, method: str, endpoint: str, data: Any | None):
        async with self.session.request(
            method,
            endpoint,
            headers={"Authorization": f"Bearer {AsyncPubsubClient.get_auth_token()}"},
            data=data,
        ) as resp:
            if resp.status != 200:
                raise ServerError(await resp.text())
            return await resp.json()

    async def pull(
        self, max_messages: int = 20, timeout: int = 10
    ) -> list[SubscriptionMessage]:
        async with asyncio.timeout(timeout):
            body = await self._make_api_request(
                "POST",
                f"{self.api_url}:pull",
                {
                    "returnImmediately": False,
                    "maxMessages": max_messages,
                },
            )

            if "receivedMessages" not in body:
                return []

            messages = []
            for message in body["receivedMessages"]:
                body = message.get("message")
                if body is not None:
                    body = PubsubMessage(
                        data=message["message"].get("data"),
                        attributes=dict(message["message"].get("attributes", {})),
                        messageId=message["message"]["messageId"],
                        publishTime=message["message"]["publishTime"],
                        orderingKey=message["message"].get("orderingKey"),
                    )
                messages.append(
                    SubscriptionMessage(
                        ackId=message.get("ackId"),
                        message=body,
                        deliveryAttempt=message.get("deliveryAttempt"),
                    )
                )
            return messages

    async def acknowledge(self, ack_ids: list[str]):
        await self._make_api_request(
            "POST",
            f"{self.api_url}:acknowledge",
            {
                "ackIds": ack_ids,
            },
        )



import asyncio
from datetime import datetime, timedelta
from http import HTTPStatus
from typing import Literal, TypeAlias
from urllib.parse import urlparse

import opentelemetry.context as context
from aiobotocore.client import AioBaseClient
from aiobotocore.config import AioConfig
from aiobotocore.session import get_session
from aiobotocore.signers import AioRequestSigner
from botocore.client import Config
from botocore.exceptions import ClientError
from botocore.session import Session
from latch_data_validation.data_validation import validate
from latch_o11y.o11y import dict_to_attrs, trace_function_with_span
from opentelemetry.sdk.resources import Attributes
from opentelemetry.trace import Span, SpanKind, Status, StatusCode, get_tracer
from types_aiobotocore_s3.client import S3Client
from types_aiobotocore_s3.type_defs import (
    EmptyResponseMetadataTypeDef,
    ResponseMetadataTypeDef,
)

tracer = get_tracer(__name__)

AWSRegion: TypeAlias = (
    Literal["us-east-1"]
    | Literal["us-east-2"]
    | Literal["us-west-1"]
    | Literal["us-west-2"]
    | Literal["af-south-1"]
    | Literal["ap-east-1"]
    | Literal["ap-southeast-1"]
    | Literal["ap-southeast-2"]
    | Literal["ap-southeast-3"]
    | Literal["ap-south-1"]
    | Literal["ap-northeast-3"]
    | Literal["ap-northeast-2"]
    | Literal["ap-northeast-1"]
    | Literal["ca-central-1"]
    | Literal["eu-central-1"]
    | Literal["eu-west-1"]
    | Literal["eu-west-2"]
    | Literal["eu-west-3"]
    | Literal["eu-south-1"]
    | Literal["eu-north-1"]
    | Literal["me-south-1"]
    | Literal["sa-east-1"]
    | Literal["us-gov-east-1"]
    | Literal["us-gov-west-1"]
)

sess = get_session()

max_presigned_url_age = timedelta(days=7) // timedelta(seconds=1)

# >>> TCP Keep-alive
# see https://github.com/boto/botocore/issues/2249 and https://github.com/boto/botocore/pull/2766
# aiobotocore uses a botocore that is really old and so does not have this in the config

_vanilla_get_scoped_config = Session.get_scoped_config


def keep_alive_get_scoped_config(self: Session):
    res = _vanilla_get_scoped_config(self)
    res["tcp_keepalive"] = True
    return res


Session.get_scoped_config = keep_alive_get_scoped_config

# >>> Tracing

_ignore_errors_key = context.create_key("_aws_ignore_errors")

_vanilla_generate_presigned_url = AioRequestSigner.generate_presigned_url

_vanilla_make_api_call = AioBaseClient._make_api_call


def add_response_metadata_attributes(span: Span, meta: ResponseMetadataTypeDef):
    status_code = meta["HTTPStatusCode"]
    if (400 <= status_code <= 499 or 500 <= status_code <= 599) and context.get_value(
        _ignore_errors_key
    ) is not True:
        span.set_status(Status(StatusCode.ERROR, HTTPStatus(status_code).phrase))

    attrs: Attributes = {
        "http.status_code": status_code,
        "rpc.aws-api.request_id": meta["RequestId"],
        "rpc.aws-api.host_id": meta["HostId"],
        # datadog
        "aws.requestid": meta["RequestId"],
    }
    if "x-amz-id-2" in meta["HTTPHeaders"]:
        attrs["rpc.aws-api.request_id_2"] = meta["HTTPHeaders"]["x-amz-id-2"]
        # datadog
        attrs["aws.requestid2"] = meta["HTTPHeaders"]["x-amz-id-2"]
    span.set_attributes(attrs)


async def traced_generate_presigned_url(
    self: AioRequestSigner,
    request_dict,
    operation_name: str,
    expires_in: int = 3600,
    region_name=None,
    signing_name=None,
):
    # todo(maximsmol): add more attributes?
    with tracer.start_as_current_span(
        f"generate_presigned_url.{operation_name}",
        attributes={} | dict_to_attrs(request_dict, "params"),
    ):
        return await _vanilla_generate_presigned_url(
            self, request_dict, operation_name, expires_in, region_name, signing_name
        )


async def traced_make_api_call(self: AioBaseClient, operation_name: str, api_params):
    # todo(maximsmol): add otel attributes
    # http.method
    # http.flavor
    # http.url
    # http.request_content_length
    # http.response_content_length
    # net.peer.port
    # http.retry_count - note: this must be in spans created *per request*
    #   the datadog instrumentation goes against spec here

    endpoint_parts = urlparse(self._endpoint.host)

    port = endpoint_parts.port
    if port is None:
        if endpoint_parts.scheme == "http":
            port = 80
        elif endpoint_parts.scheme == "https":
            port = 443

    endpoint_name = self._endpoint._endpoint_prefix
    with tracer.start_as_current_span(
        f"aws-api.{endpoint_name}/{operation_name}",
        # todo(maximsmol): remove if not needed
        # f"{endpoint_name}.command", # dd span name (kept around in case dd won't recognize the otel one)
        kind=SpanKind.CLIENT,
        attributes={
            # todo(maximsmol): is this service name override a datadog idiosyncrasy?
            "service.name": f"aws.{endpoint_name}",
            "http.scheme": endpoint_parts.scheme,
            "http.host": endpoint_parts.netloc,
            "net.peer.name": str(endpoint_parts.hostname),
            "rpc.method": operation_name,
            "rpc.system": "aws-api",
            "rpc.service": endpoint_name,
            # datadog:
            "span.type": "http",
            "resource.name": f"{endpoint_name}.{operation_name.lower()}",
            "aws.agent": "aiobotocore",
            "aws.operation": operation_name,
            "aws.region": self.meta.region_name,
        }
        | (
            {
                "net.peer.port": port,
            }
            if port is not None
            else {}
        )
        | dict_to_attrs(api_params, "rpc.aws-api.params"),
        set_status_on_exception=context.get_value(_ignore_errors_key) is not True,
    ) as s:
        try:
            res: EmptyResponseMetadataTypeDef = await _vanilla_make_api_call(
                self, operation_name, api_params
            )

            add_response_metadata_attributes(s, res["ResponseMetadata"])

            return res
        except ClientError as e:
            err_res = e.response

            attrs: Attributes = {}
            if "Error" in err_res:
                err = err_res["Error"]
                if "Code" in err:
                    attrs["rpc.aws-api.error_code"] = err["Code"]
                if "Message" in err:
                    attrs["rpc.aws-api.error_message"] = err["Message"]
            if "Status" in err_res:
                attrs["rpc.aws-api.status"] = err_res["Status"]
            if "StatusReason" in err_res:
                attrs["rpc.aws-api.status_reason"] = err_res["StatusReason"]
            s.set_attributes(attrs)

            if "ResponseMetadata" in err_res:
                add_response_metadata_attributes(s, err_res["ResponseMetadata"])

            raise e


AioRequestSigner.generate_presigned_url = traced_generate_presigned_url
AioBaseClient._make_api_call = traced_make_api_call

# >>> Main

_s3_clients: dict[AWSRegion, S3Client] = {}


@trace_function_with_span(tracer)
async def s3_client(s: Span, region: AWSRegion = "us-west-2"):
    s.set_attributes({"rpc.aws-api.region": region, "cached": region in _s3_clients})

    if region not in _s3_clients:
        _s3_clients[region] = await sess.create_client(
            "s3",
            region_name=region,
            config=AioConfig().merge(Config(signature_version="s3v4")),
        ).__aenter__()

    return _s3_clients[region]


async def s3_cleanup():
    async def close_client(region: AWSRegion, x: S3Client):
        with tracer.start_as_current_span(
            "close s3 client", attributes={"rpc.aws-api.region": region}
        ):
            await x.__aexit__(None, None, None)

    deferred: list[asyncio.Task[object]] = []
    for k, v in _s3_clients.items():
        deferred.append(asyncio.create_task(close_client(k, v)))

    await asyncio.gather(*deferred)


_bucket_region_cache: dict[str, tuple[AWSRegion, datetime]] = {}


@trace_function_with_span(tracer)
async def get_bucket_region(
    s: Span,
    *,
    Bucket: str,  # noqa: N803 (variables should be lowercase)
    ExpectedBucketOwner: str | None = None,  # noqa: N803
) -> AWSRegion:
    now = datetime.now()

    cached_value = _bucket_region_cache.get(Bucket)
    if cached_value is not None and cached_value[1] <= now:
        cached_value = None
        del _bucket_region_cache[Bucket]

    s.set_attributes({"bucket": Bucket, "cached": cached_value is not None})
    if cached_value is not None:
        return cached_value[0]

    default_s3 = await s3_client()

    ctx_reset_token: object | None = None
    try:
        new_ctx = context.set_value(_ignore_errors_key, True)
        ctx_reset_token = context.attach(new_ctx)

        if ExpectedBucketOwner is not None:
            head = await default_s3.head_bucket(
                Bucket=Bucket, ExpectedBucketOwner=ExpectedBucketOwner
            )
        else:
            head = await default_s3.head_bucket(Bucket=Bucket)

        region = head["ResponseMetadata"]["HTTPHeaders"]["x-amz-bucket-region"]
    except ClientError as e:
        # the request WILL error for some reason despite returning the correct header

        if e.operation_name != "HeadBucket":
            raise e

        region = (
            e.response.get("ResponseMetadata", {})
            .get("HTTPHeaders", {})
            .get("x-amz-bucket-region")
        )
        if region is None:
            raise e
    finally:
        if ctx_reset_token is not None:
            context.detach(ctx_reset_token)

    res: AWSRegion = validate(region, AWSRegion)
    _bucket_region_cache[Bucket] = (res, now + timedelta(hours=1))
    return res



from dataclasses import dataclass
from datetime import datetime, timedelta
from typing import Any, TypedDict
from urllib.parse import urljoin
from xml.etree import ElementTree as ET

import aiohttp
import xmltodict
from dateutil.parser import parse
from google.api_core.exceptions import (
    BadRequest,
    Forbidden,
    NotFound,
    ServerError,
    TooManyRequests,
    Unauthorized,
)
from google.auth import default
from google.auth.transport import requests
from google.cloud import storage

max_presigned_url_age = timedelta(days=7) // timedelta(seconds=1)


@dataclass(frozen=True)
class Bucket:
    name: str
    id: str
    creation_time: str
    update_time: str
    project_number: str


@dataclass(frozen=True)
class Object:
    id: str
    name: str
    bucket: str
    media_link: str
    generation: str
    content_type: str
    size: int
    creation_time: datetime
    update_time: datetime


CompletedPartTypeDef = TypedDict(
    "CompletedPartTypeDef",
    {
        "ETag": str,
        "ChecksumCRC32": str,
        "ChecksumCRC32C": str,
        "ChecksumSHA1": str,
        "ChecksumSHA256": str,
        "PartNumber": int,
    },
    total=False,
)


class AsyncHTTPIterator:
    def __init__(self, url: str, headers: dict, params: dict, storage_client):
        self.url = url
        self.headers = headers
        self.params = params
        self.next_page_token = None
        self.prefixes = list()
        self.items = []
        self.storage_client = storage_client

    def __aiter__(self):
        return self

    async def __anext__(self):
        if not self.items and self.next_page_token == "":
            raise StopAsyncIteration

        if self.items:
            obj = self.items.pop()
            return Object(
                id=obj["id"],
                name=obj["name"],
                bucket=obj["bucket"],
                media_link=obj["mediaLink"],
                generation=obj["generation"],
                content_type=obj["contentType"],
                size=int(obj["size"]),
                creation_time=parse(obj["timeCreated"]),
                update_time=parse(obj["updated"]),
            )

        if self.next_page_token:
            self.params["pageToken"] = self.next_page_token

        res = await self.storage_client._make_api_request(
            "GET", self.url, headers=self.headers, params=self.params, return_json=True
        )

        if "items" not in res:
            raise StopAsyncIteration

        self.items = res.get("items", [])
        self.prefixes.extend(res.get("prefixes", []))
        self.next_page_token = res.get("nextPageToken", "")

        return await self.__anext__()


class AsyncStorageClient:
    credentials, _ = default(scopes=["https://www.googleapis.com/auth/cloud-platform"])

    def __init__(self, session: aiohttp.ClientSession):
        self.api_url = "https://storage.googleapis.com/storage/v1/"
        self.upload_api_url = "https://storage.googleapis.com/upload/storage/v1/"
        self.session = session

    @staticmethod
    def get_auth_token():
        credentials = AsyncStorageClient.credentials

        if credentials.token is None or credentials.expired:
            credentials.refresh(requests.Request())
        return credentials.token

    async def _make_api_request(
        self,
        method: str,
        url: str,
        *,
        data: Any | None = None,
        headers: dict | None = None,
        params: dict | None = None,
        return_bytes=False,
        return_json=False,
        return_response=False,
    ) -> Any:
        headers = headers or {}
        headers["Authorization"] = f"Bearer {AsyncStorageClient.get_auth_token()}"

        async with self.session.request(
            method,
            url,
            headers=headers,
            params=params,
            data=data,
        ) as resp:
            if resp.status >= 400:
                err = await resp.text()
                if resp.status == 400:
                    raise BadRequest(message=err, response=resp)
                if resp.status == 401:
                    raise Unauthorized(message=err, response=resp)
                if resp.status == 403:
                    raise Forbidden(message=err, response=resp)
                if resp.status == 404:
                    raise NotFound(message=err, response=resp)
                if resp.status == 429:
                    raise TooManyRequests(message=err, response=resp)
                raise ServerError(message=err, response=resp)
            if resp.status == 204:
                return
            if return_bytes:
                return await resp.content.read()
            if return_json:
                return await resp.json()
            if return_response:
                return resp
            return await resp.text()

    async def get_bucket(self, bucket_name: str):
        try:
            res = await self._make_api_request(
                "GET",
                urljoin(
                    self.api_url,
                    f"b/{bucket_name}",
                ),
                return_json=True,
            )
        except NotFound:
            return None

        return Bucket(
            name=res["name"],
            id=res["id"],
            creation_time=res["timeCreated"],
            update_time=res["updated"],
            project_number=res["projectNumber"],
        )

    async def get_blob_meta(self, bucket_name: str, key: str):
        try:
            res = await self._make_api_request(
                "GET",
                urljoin(self.api_url, f"b/{bucket_name}/o/{key}"),
                return_json=True,
            )
        except NotFound:
            return None

        return Object(
            name=res["name"],
            id=res["id"],
            bucket=res["bucket"],
            media_link=res["mediaLink"],
            generation=res["generation"],
            content_type=res["contentType"],
            size=int(res["size"]),
            creation_time=parse(res["timeCreated"]),
            update_time=parse(res["updated"]),
        )

    def list_blobs(
        self,
        bucket_name: str,
        prefix: str | None = None,
        delimiter: str = "/",
        match_glob: str | None = None,
        page_size: int = 1000,
        page_token: str | None = None,
        include_trailing_delimeter: bool | None = None,
        include_folders_as_prefixes: bool | None = None,
    ):
        if match_glob is not None and delimiter != "/":
            raise ValueError("match_glob is only supported with delimeter='/'")

        if include_folders_as_prefixes and delimiter != "/":
            raise ValueError(
                "include_folders_as_prefixes is only supported with delimeter='/'"
            )

        params = {
            "prefix": prefix,
            "delimiter": delimiter,
            "matchGlob": match_glob,
            "maxResults": page_size,
            "pageToken": page_token,
            "includeTrailingDelimeter": include_trailing_delimeter,
            "includeFoldersAsPrefixes": include_folders_as_prefixes,
        }

        params = {k: str(v) for k, v in params.items() if v is not None}

        return AsyncHTTPIterator(
            url=urljoin(self.api_url, f"b/{bucket_name}/o"),
            headers={},
            params=params,
            storage_client=self,
        )

    async def get_blob_bytes(
        self,
        bucket_name: str,
        key: str,
        start: int | None = None,
        end: int | None = None,
    ):
        headers = {}
        start = start if start is not None else 0

        if end is not None:
            headers["Range"] = f"bytes={start}-{end}"

        return bytes(
            await self._make_api_request(
                "GET",
                urljoin(self.api_url, f"b/{bucket_name}/o/{key}"),
                params={"alt": "media"},
                headers=headers,
                return_bytes=True,
            )
        )

    async def put_blob(
        self,
        bucket_name: str,
        key: str,
        data: bytes,
        content_type: str | None = None,
    ):
        if content_type is None:
            content_type = "text/plain"

        resp = await self._make_api_request(
            "POST",
            urljoin(self.upload_api_url, f"b/{bucket_name}/o"),
            params={"name": key, "uploadType": "media"},
            headers={"Content-Type": content_type},
            data=data,
            return_json=True,
        )

        return Object(
            id=resp["id"],
            name=resp["name"],
            bucket=resp["bucket"],
            media_link=resp["mediaLink"],
            generation=resp["generation"],
            content_type=resp["contentType"],
            size=int(resp["size"]),
            creation_time=parse(resp["timeCreated"]),
            update_time=parse(resp["updated"]),
        )

    async def copy_blob(
        self,
        src_bucket_name: str,
        src_bucket_key: str,
        dest_bucket_name: str,
        dest_bucket_key: str,
    ):
        done = False
        rewrite_token = None
        try:
            while not done:
                params = {}
                if rewrite_token:
                    params = {"rewriteToken": rewrite_token}

                resp = await self._make_api_request(
                    "POST",
                    urljoin(
                        self.api_url,
                        f"b/{src_bucket_name}/o/{src_bucket_key}/rewriteTo/b/{dest_bucket_name}/o/{dest_bucket_key}",
                    ),
                    params=params,
                    headers={"Content-Type": "application/json"},
                    return_json=True,
                )

                done = resp["done"]
                rewrite_token = resp.get("rewriteToken", None)
        except NotFound:
            return None

    async def delete_blob(
        self, bucket_name: str, key: str, generation: str | None = None
    ):
        params = {}
        if generation is not None:
            params["generation"] = generation

        await self._make_api_request(
            "DELETE",
            urljoin(self.api_url, f"b/{bucket_name}/o/{key}"),
            params=params,
        )

    def get_signed_url(
        self,
        bucket_name: str,
        key: str | None,
        content_disposition: str | None = None,
        content_type: str | None = None,
    ):
        # note(taras): this constructs a blob object and signs download url without making requests
        blob = storage.Client().bucket(bucket_name).blob(key)

        return blob.generate_signed_url(
            version="v4",
            expiration=max_presigned_url_age,
            method="GET",
            content_type=content_type,
            response_disposition=content_disposition,
        )

    async def get_signed_upload_url(
        self,
        bucket_name: str,
        key: str,
        upload_id: str,
        part_number: int,
    ):

        meta = await self.get_blob_meta(bucket_name, key)
        if meta is None:
            await self.put_blob(bucket_name, key, b"")

        # note(taras): this constructs a blob object and signs
        # upload url without making syncronous requests
        blob = storage.Client().bucket(bucket_name).blob(key)

        return blob.generate_signed_url(
            version="v4",
            expiration=max_presigned_url_age,
            method="PUT",
            query_parameters={
                "partNumber": part_number,
                "uploadId": upload_id,
            },
        )

    async def multipart_initiate_upload(
        self,
        bucket_name: str,
        bucket_key: str,
        content_type: str = "text/plain",
    ):
        # note(taras): multipart uploads use XML api instead of JSON
        resp = await self._make_api_request(
            "POST",
            f"https://storage.googleapis.com/{bucket_name}/{bucket_key}?uploads",
            headers={
                "Content-Type": content_type,
            },
        )

        return str(xmltodict.parse(resp)["InitiateMultipartUploadResult"]["UploadId"])

    async def multipart_upload_part(
        self,
        bucket_name: str,
        bucket_key: str,
        upload_id: str,
        part_number: int,
        data: bytes,
    ):
        resp = await self._make_api_request(
            "PUT",
            f"https://storage.googleapis.com/{bucket_name}/{bucket_key}?uploadId={upload_id}&partNumber={part_number}",
            data=data,
            return_response=True,
        )

        return resp.headers["ETag"]

    async def multipart_complete_upload(
        self,
        bucket_name: str,
        bucket_key: str,
        upload_id: str,
        parts: list[CompletedPartTypeDef],
    ):
        root = ET.Element("CompleteMultipartUpload")

        for part in parts:
            if "PartNumber" not in part:
                raise ValueError("PartNumber is required")
            if "ETag" not in part:
                raise ValueError("ETag is required")

            part_element = ET.SubElement(root, "Part")
            ET.SubElement(part_element, "PartNumber").text = str(part["PartNumber"])
            ET.SubElement(part_element, "ETag").text = part["ETag"]

        await self._make_api_request(
            "POST",
            f"https://storage.googleapis.com/{bucket_name}/{bucket_key}?uploadId={upload_id}",
            data=ET.tostring(root),
        )

        blob = await self.get_blob_meta(bucket_name, bucket_key)
        if blob is None:
            raise ServerError("Failed to complete multipart upload. Can't find blob")

        return blob.generation
